/*
 * connecting_screen.cpp
 *
 * Copyright (c) Stephen Thompson, 2007.
 * Licensed for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "cfg.hpp"
#include "connecting_screen.hpp"
#include "loading_screen.hpp"
#include "display.hpp"
#include "x_centre.hpp"

#ifdef _WIN32_WCE
#include "wce_serial.hpp"
#else
#include "enet_streams/enet_controller.hpp"
#include "enet_streams/enet_stream.hpp"
#endif

#include "boost/scoped_ptr.hpp"
using namespace boost;

namespace {
	ConnectingScreen *conscr = 0;
#ifndef _WIN32_WCE
	scoped_ptr<EnetController> enet_controller; // this stays alive until the end of the program.
#endif
}

void SetupConnectingScreen(const UsefulStuff &u)
{
	ASSERT(!conscr);
	conscr = new ConnectingScreen(u);
	conscr->render();
	u.app2d->addObserver(conscr);
	u.app2d->setUpdateIntervalActive(50);
	u.app2d->setUpdateIntervalInactive(50);
	u.app2d->setRenderFast(false);
}

void StopConnectingScreen(const UsefulStuff &u)
{
	if (!conscr) return;
	u.app2d->rmObserver(conscr);
	delete conscr;
	conscr = 0;
}

ConnectingScreen::ConnectingScreen(const UsefulStuff &u)
	: us(u)
{
#ifdef _WIN32_WCE
	// Open the COM port
	if (us.server) {
		wce_str.reset(new WceSerialStream(CfgInt("in_com")));
	} else {
		wce_str.reset(new WceSerialStream(CfgInt("out_com")));
	}
	us.remote = wce_str;
#else
	// Start up enet controller, and start up server if appropriate
	enet_controller.reset(new EnetController);
	if (us.server) {
		enet_controller->startServer(CfgInt("port_number"));
	}
#endif
}

void ConnectingScreen::render()
{
    us.display->setFontSize(CfgInt("font_size"));
	us.display->begin();
	XCentre(*us.display, us.display->getTextHeight() + 15, "CONNECTING");
	us.display->end();
}

void ConnectingScreen::update()
{
	bool done = false;
#ifdef _WIN32_WCE
	done = wce_str->tryConnect();
#else
	if (us.server) {
		// Listen for clients connecting to the server
		if (clients.empty()) {
			enet_controller->getClients(clients);
			enet_controller->processMessages(1000); // allow enet to block for a bit.
		}
		if (!clients.empty()) {
			// Somebody has connected -- we are done. 
			us.remote = clients.front();
			done = true;
		}
	} else {
		// Connect to the server; time out if the server does not respond within 5 sec.
		const string server_name(CfgString("server_name"));
		const int port_number(CfgInt("port_number"));
		us.remote = enet_controller->connectToServer(server_name.c_str(), port_number, 5000);
		if (!us.remote) exit(1); // (TODO) -- better error handling...
		done = true;
	}
#endif

	if (done) {
		// Enable exceptions on the remote stream. This should catch unexpected EOF errors.
		us.remote->clear();
		us.remote->exceptions(ios_base::badbit | ios_base::failbit | ios_base::eofbit);
		
		// Go to the loading screen next.
		SetupLoadingScreen(us);
		StopConnectingScreen(us);
	}
}

