/* Utility routine to load bitmaps */
/* Based on code by Jeff Lait, used under BSD licence. Modified by Stephen Thompson. See ACKNOWLEDGMENTS.TXT. */

#include "misc.hpp"

#include "loadbitmap.hpp"
#include "my_exceptions.hpp"

#include <istream>
using namespace std;

typedef struct
{
    char    id[2];
    int	    size;
    int reserved;
    int	    headersize;
    int	    infosize;
    int	    width;
    int	    depth;
    short   biplane;
    short   bits;
    int	    comp;
    int	    imagesize;
    int	    xpelperm;
    int	    ypelperm;
    int	    clrused;
    int	    clrimport;
} BMPHEAD;

typedef struct
{
    unsigned char blue;
    unsigned char green;
    unsigned char red;
    unsigned char filler;
} RGBQUAD;



/* read little-endian short */
static short
gfx_readshort(istream &str)
{
    unsigned char c[2] = {0,0};
    str.read((char*)c, 2);
    return (c[0] + (c[1] << 8));
}

/* read little-endian long */
static int
gfx_readint(istream &str)
{
    unsigned char c[4] = {0,0,0,0};
    str.read((char*)c, 4);
    return (c[0] + (c[1] << 8) + (c[2] << 16) + (c[3] << 24));
}

shared_array<RGB> gfx_loadbitmap(istream &str, int *ww, int *hh)
{
    shared_array<RGB> result;

    BMPHEAD head;
    RGBQUAD palette[256];

    int i, x, y, w, h;

    if (!str) throw GraphicLoadFailed("could not access stream");

    head.id[0] = head.id[1] = 0;
    str.read(head.id, 2);

    head.size = gfx_readint(str);
    head.reserved = gfx_readint(str);
    head.headersize = gfx_readint(str);
    head.infosize = gfx_readint(str);
    head.width = gfx_readint(str);
    head.depth = gfx_readint(str);
    head.biplane = gfx_readshort(str);
    head.bits = gfx_readshort(str);
    head.comp = gfx_readint(str);
    head.imagesize = gfx_readint(str);
    head.xpelperm = gfx_readint(str);
    head.ypelperm = gfx_readint(str);
    head.clrused = gfx_readint(str);
    head.clrimport = gfx_readint(str);

    if (head.id[0] != 'B' || head.id[1] != 'M') throw GraphicLoadFailed("not a BMP file");

    if (head.bits != 24 && head.bits != 8) throw GraphicLoadFailed("only 8-bit or 24-bit BMPs are supported");

    h = head.depth;
    w = head.width;
    result.reset(new RGB [h * w]);
    
    if (head.bits == 8)
    {
        str.read((char*)palette, 256*sizeof(RGBQUAD));
    }

    /* Work out the number of padding bytes at the end of each row. */
    const int bytes_per_row = (head.bits == 24) ? w*3 : w;
    const int left_over = bytes_per_row % 4;
    const int padding = (left_over == 0) ? 0 : 4 - left_over;

    /* BMPs are stored bottom up, so we reverse it here. */
    for (y = head.depth - 1; y >= 0; y--)
    {
        for (x = 0; x < head.width; x++)
        {
            unsigned char red, blue, green, idx;

            if (head.bits == 24)
            {
                /* BMPs are stored BGR */
                str.read((char*)&blue, 1);
                str.read((char*)&green, 1);
                str.read((char*)&red, 1);
            }
            else
            {
                str.read((char*)&idx, 1);
                red = palette[idx].red;
                green = palette[idx].green;
                blue = palette[idx].blue;
            }

            i = x + y * w;
            
            result[i].r = red;
            result[i].g = green;
            result[i].b = blue;
        }

        /* Account for padding to DWORD boundary */
        if (padding > 0) str.read((char*)&i, padding);  /* i is being used as a dummy variable here. */
    }

    if (ww) *ww = w;
    if (hh) *hh = h;

    return result;
}
