/*
 * menu_screen.cpp
 *
 * Copyright (c) Stephen Thompson, 2007.
 * Licensed for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "cfg.hpp"
#include "dungeon_generator.hpp"
#include "in_game_screen.hpp"
#include "menu.hpp"
#include "menu_screen.hpp"
#include "menu_selections.hpp"
#include "my_exceptions.hpp"
#include "display.hpp"
#include "graphic.hpp"
#include "round.hpp"
#include "server_config.hpp"
#include "x_centre.hpp"

#include <sstream>
using namespace std;

namespace {
    const int extra_spacing = 2;   // extra spacing between menu lines (in pixels).
}

namespace {
    MenuScreen *mscr = 0;
}

MenuSelections MenuScreen::menu_selections;
bool MenuScreen::first_time = true;

void SetupMenuScreen(const UsefulStuff &u)
{
    ASSERT(!mscr);
    mscr = new MenuScreen(u);
    mscr->render();
    u.app2d->addObserver(mscr);
    u.app2d->setUpdateIntervalActive(250);
    u.app2d->setUpdateIntervalInactive(250);
    u.app2d->setRenderFast(false);
}

void StopMenuScreen(const UsefulStuff &u)
{
    if (!mscr) return;
    u.app2d->rmObserver(mscr);
    delete mscr;
    mscr = 0;
}

namespace {
    int RenderMenuValue(Display &dpy, int val, const MenuValue &mval, int x, int y, bool draw)
    {
        // If draw==true, draws the menu value and returns its width
        // Else, just returns its width

        // Note -- coords referred to bottom left of display

        static const int menu_gfx_percent = CfgInt("menu_gfx_percent");
        
        int width = 0;
        
        if (!mval.getText().empty()) {
            const char *txt = mval.getText().c_str();
            width = dpy.getTextWidth(txt);
            if (draw) dpy.drawText(x, y, txt);
        } else if (mval.getCharIcon()) {
            char txt[2];
            txt[0] = mval.getCharIcon();
            txt[1] = 0;
            const int w = dpy.getTextWidth(txt);
            width = w*val;
            if (draw) {
                for (int i=0; i<val; ++i) {
                    dpy.drawText(x, y, txt);
                    x += w;
                }
            }
        } else if (mval.getGraphicIcon()) {
            // Scale menu graphics before drawing
            const float scale = float(menu_gfx_percent)/100.0f;
            float round_up, round_down;
            dpy.getGfxResizer()->roundScaleFactor(scale, round_down, round_up);
            const int h = Round(GetHeight(*mval.getGraphicIcon()) * round_down);
            const int w = Round(GetWidth(*mval.getGraphicIcon()) * round_down);
            const int yadd = max(0, (dpy.getTextHeight() - h)/2);
            width = w*val;
            if (draw) {
                for (int i=0; i<val; ++i) {
                    dpy.drawGraphic(x, y + yadd, *mval.getGraphicIcon(), w, h);
                    x += w;
                }
            }
        } else {
            // Assume that we want it drawn as a number.
            ostringstream oss;
            oss << val;
            width = dpy.getTextWidth(oss.str().c_str());
            if (draw) {
                dpy.drawText(x, y, oss.str().c_str());
            }
        }

        return width;
    }
        
    int CalcNumLines(Display &dpy, const MenuItem &item)
    {
        const string txt = (item.getTitle() + ": ");
        const int w1 = dpy.getTextWidth(txt.c_str());
        int w2 = 0;
        
        for (int i=item.getMinValue(); i < item.getMinValue()+item.getNumValues(); ++i) {
            const int w = RenderMenuValue(dpy, i, item.getValue(i), 0, 0, false);
            if (w > w2) w2 = w;
        }

        if (w1 + w2 > dpy.getWidth()) {
            // Item too wide; split it out into 2 lines
            return 2;
        } else {
            // Item can fit onto one line (whatever its value is)
            return 1;
        }
    }
}

MenuScreen::MenuScreen(const UsefulStuff &u)
    : us(u), current_page(0), lines_per_page(1)
{
    if (us.server) {

        setupPages();
        
        // Set up "menu_selections" if this is the first run through.
        if (first_time) {
            for (int i=0; i<us.server_config->getMenu().getNumItems(); ++i) {
                const MenuItem &item(us.server_config->getMenu().getItem(i));

                // Items are set to their min_value initially, except for "quest",
                // which is set to 1 (the first "real" quest value -- not CUSTOM).
                if (item.getID() == "quest") {
                    menu_selections.set("quest", 1);
                } else {
                    menu_selections.set(item.getID(), item.getMinValue());
                }
            }

        }

        first_time = false;
        
        // Ensure the constraints are applied properly...
        applyMenuConstraints();
    }
}

void MenuScreen::setupPages()
{
    if (us.server) {
        // Divide the menu items into pages
        nlines.resize(us.server_config->getMenu().getNumItems());
        
        int line_count = 0;
        lines_per_page = us.display->getHeight() / (us.display->getTextHeight() + extra_spacing) - 1;
        if (lines_per_page < 1) lines_per_page = 1;
        
        pages.clear();
        pages.push_back(0);
        
        for (int i=0; i<us.server_config->getMenu().getNumItems(); ++i) {    
            const MenuItem &item(us.server_config->getMenu().getItem(i));        
            
            nlines[i] = CalcNumLines(*us.display, item);
            line_count += nlines[i];
            if (line_count > lines_per_page) {
                pages.push_back(i); // This item should start on a new page
                line_count = nlines[i];
            }
            
        }        
        pages.push_back(us.server_config->getMenu().getNumItems());

        const int mlpp = CfgInt("menu_lines_per_page");
        if (pages.size()==2 && line_count <= mlpp && lines_per_page > mlpp) {
            lines_per_page = mlpp;
        }
        
        // reset current_page if needed.
        if (current_page >= pages.size()-1) current_page = 0;
    }
}

void MenuScreen::render()
{
    us.display->setFontSize(CfgInt("font_size"));
    us.display->begin(0, 0, 0);

    if (us.server) {

        const Menu & menu(us.server_config->getMenu());

        const int start_item = pages[current_page];
        const int finish_item = pages[current_page+1];

        const int text_height = us.display->getTextHeight() + extra_spacing;
        
        int ypos = text_height;
        
        for (int i = start_item; i < finish_item; ++i) {
            const MenuItem &item(menu.getItem(i));

            // Draw the item's title string
            string title = item.getTitle() + ": ";
            us.display->drawText(0, us.display->getHeight() - 1 - ypos,
                                 title.c_str());

            // Draw the value 
            const int val = menu_selections.get(menu.getItem(i).getID());
            const MenuValue &mval(menu.getItem(i).getValue(val));

            if (nlines[i] == 2) {
                RenderMenuValue(*us.display, val, mval, 30,
                                us.display->getHeight() - 1 - (ypos+text_height), true);
            } else {
                const int w = us.display->getTextWidth(title.c_str());
                RenderMenuValue(*us.display, val, mval, w,
                                us.display->getHeight() - 1 - ypos, true);
            }

            ypos += nlines[i]*text_height;
        }

        ypos = (1 + lines_per_page)*text_height;
        us.display->drawText(0, us.display->getHeight() - 1 - ypos, "START");
        if (pages.size()>2) {
            us.display->drawText(us.display->getWidth() - us.display->getTextWidth("PAGE"),
                                 us.display->getHeight() - 1 - ypos, "PAGE");
        }
        
    } else {
        // The menu is not displayed on the client ...
        XCentre(*us.display, us.display->getTextHeight() + 15, "WAITING FOR");
        XCentre(*us.display, 2*us.display->getTextHeight() + 20, "GAME TO BEGIN");
    }
    
    us.display->end();
}

void MenuScreen::update()
{
    if (!us.server && us.remote) {
        // When we receive the 'M', we can leave the menu...
        if (us.remote->rdbuf()->in_avail() > 0) {
            char c = us.remote->get();
            if (c != 'M') throw UnexpectedError("network protocol error");
            SetupInGameScreen(us, menu_selections);
            StopMenuScreen(us);
        }
    }
}           

void MenuScreen::mouseDown(int x, int y, bool rmb)
{
    if (!us.server) return;
    
    int line = y / (us.display->getTextHeight() + extra_spacing);
    if (line < lines_per_page) {
        // Find which item
        int item = pages[current_page];
        while (line >= nlines[item]) {
            line -= nlines[item];
            ++item;
        }

        if (item < us.server_config->getMenu().getNumItems()) {
        
            // Increase value for that item (or decrease if RMB was clicked)
            const MenuItem &mit(us.server_config->getMenu().getItem(item));
            const string &id(mit.getID());
            const int cur_val = menu_selections.get(id);
            const int min_val = mit.getMinValue();
            const int max_val = mit.getMinValue() + mit.getNumValues() - 1;
            int new_val;
            
            if (rmb) {
                new_val = cur_val - 1;
                if (new_val < min_val) {
                    new_val = min_val;
                }
                // Special case for "Quest" item -- skip over "CUSTOM" (min. value)
                if (new_val == min_val && mit.getID() == "quest") {
                    new_val = min_val + 1;
                }
            } else {
                new_val = cur_val + 1;
                if (new_val > max_val) {
                    new_val = min_val;
                    if (mit.getID() == "quest") {
                        // Special case -- skip over "CUSTOM" on the "Quest" item
                        ++new_val;
                    }
                }
            }
            
            if (mit.getID() != "quest") {
                // Special case -- when a non-Quest item is clicked, we reset the
                // Quest to "CUSTOM"
                menu_selections.set("quest", 0);
            }
            menu_selections.set(id, new_val);
            applyMenuConstraints();

            if (menu_selections.get(id) == cur_val) {
                // oops, nothing changed
                // try going back to the min (if LMB) or max (if RMB) instead
                menu_selections.set(id, rmb? max_val : min_val);
                applyMenuConstraints();
            }
            render();

        }

    } else {
        // In the Start/Page area
        if (!rmb) {
            if (x < us.display->getWidth() / 2) {
                // Start Game
                // Clear the screen first
                us.display->begin(); us.display->end();
                
                // Try to start the InGameScreen. This can sometimes fail (if there are
                // dungeon generation errors)
                bool ok_to_start = true;
                try {
                    SetupInGameScreen(us, menu_selections);
                } catch (const DungeonGenerationFailed&) {
                    ok_to_start = false;
                }
                if (ok_to_start) {
                    StopMenuScreen(us);
                }
            } else {
                // Flip Page
                current_page++;
                if (current_page == pages.size()-1) current_page = 0;
                render();
            }
        }
    }
}

void MenuScreen::applyMenuConstraints()
{
    if (us.server) {
        const Menu &menu(us.server_config->getMenu());

        // limit of 100 iterations to prevent infinite loops
        for (int count = 0; count < 100; ++count) {
            bool changed = false;
            for (int i=0; i<menu.getNumItems(); ++i) {
                const MenuItem &item(menu.getItem(i));
                const string &id(item.getID());
                const int cur_val = menu_selections.get(id);
                if (cur_val < item.getMinValue() || cur_val >= item.getMinValue()
                + item.getNumValues()) {
                    menu_selections.set(id, item.getMinValue());
                    changed = true;
                } else {
                    const MenuValue &val(menu.getItem(i).getValue(menu_selections.get(id)));
                    if (val.applyConstraints(menu_selections)) {
                        changed = true;
                    }
                }
            }
            if (!changed) break;
        }

    }
}

void MenuScreen::windowResize(int, int)
{
    if (us.server) {  // menu is only displayed on server.
        setupPages();
        render();
    }
}
