/*
 * sdl_display.cpp
 *
 * Copyright (c) Stephen Thompson, 2007.
 * Licensed for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "cfg.hpp"
#include "display.hpp"
#include "my_exceptions.hpp"
#include "rstream/rstream_rwops.hpp"
#include "sdl_bitmap_font.hpp"
#include "sdl_graphic.hpp"
#include "sdl_graphic_cache.hpp"
#include "sdl_pixels.hpp"
#include "sdl_truetype_font.hpp"

#include "boost/scoped_ptr.hpp"
#include <SDL.h>



//
// DisplayImpl struct for SDL
//

struct DisplayImpl {
    DisplayImpl();  // loads font, sets "screen" to SDL_GetVideoSurface() (for now)

    SDL_Surface *screen;
    SDL_Surface *font_surface;
    scoped_ptr<SDLFont> font;
    int current_font_size;
};


//
// Display class for SDL
//

void Display::construct()
{
    pimpl.reset(new DisplayImpl);
    g_sdl_graphic_cache.setDisplay(*this);
}

void Display::onChangeSizer()
{
    g_sdl_graphic_cache.clear();
}

//
// begin & end drawing
//

void Display::begin(int r, int g, int b)
{
    SDL_Surface *scr = SDL_GetVideoSurface();
    SDL_SetClipRect(scr, 0);
    SDL_FillRect(scr, 0, SDL_MapRGB(scr->format, r, g, b));
}

void Display::end()
{
    SDL_Flip(SDL_GetVideoSurface());
}

//
// get width/height
//

int Display::getWidth() const
{
    return pimpl->screen->w;
}

int Display::getHeight() const
{
    return pimpl->screen->h;
}


//
// Clipping
//

void Display::setClipRectangle(int bl_x, int bl_y, int tr_x, int tr_y)
{
    SDL_Rect clip;
    clip.x = bl_x;
    clip.y = pimpl->screen->h - tr_y;  // need to give *top* for SDL
    clip.w = tr_x - bl_x;
    clip.h = tr_y - bl_y;
    SDL_SetClipRect(pimpl->screen, &clip);
}

void Display::clearClipRectangle()
{
    SDL_SetClipRect(pimpl->screen, 0);
}

//
// Font/Text Display
//

DisplayImpl::DisplayImpl()
{
    current_font_size = -999;
    
    // Set "screen" to current display surface
    screen = SDL_GetVideoSurface();
}

void Display::setFontSize(int size)
{
    if (pimpl->current_font_size == size) return;
    pimpl->current_font_size = size;
    pimpl->font.reset(LoadFirstAvailableFont(size));
}

void Display::drawText(int bl_x, int bl_y, const char *str)
{
    // NOTE: We rely on them calling setFontSize before they call drawText
    // (otherwise nothing will get drawn, because pimpl->font will be zero).
    
    if (pimpl->font) {
        int w, h;
        pimpl->font->getTextSize(str, w, h);
        pimpl->font->drawText(pimpl->screen,
                              bl_x,
                              pimpl->screen->h - bl_y - h - 1,
                              str,
                              255, 255, 255);
    }
}

void Display::getTextSize(const char *str, int &w, int &h)
{
    if (pimpl->font) {
        pimpl->font->getTextSize(str, w, h);
    } else {
        w = h = 0;
    }
}

int Display::getTextHeight()
{
    if (pimpl->font) {
        return pimpl->font->getTextHeight();
    } else {
        return 0;
    }
}


//
// draw blended rectangle
//

void Display::drawBlendedRectangle(int bl_x, int bl_y, int width, int height, int r, int g, int b, int a)
{
    SDL_Color col = {r,g,b};
    SDL_Surface * temp_surface = SDL_CreateRGBSurface(SDL_SRCALPHA, width, height, 8, 0, 0, 0, 0);
    SDL_SetColors(temp_surface, &col, 0, 1);
    SDL_FillRect(temp_surface, 0, 0); // clear the surface to zero.
    SDL_SetAlpha(temp_surface, SDL_SRCALPHA, a);

    SDL_Rect dest_rect;
    dest_rect.x = bl_x;
    dest_rect.y = pimpl->screen->h - height - bl_y - 1;
    SDL_BlitSurface(temp_surface, 0, pimpl->screen, &dest_rect);

    SDL_FreeSurface(temp_surface);
}

//
// "Graphic" drawing
// This is just delegated to Graphic.
//

void Display::drawGraphic(int bl_x, int bl_y, const Graphic &gfx, int width, int height)
{
    const Graphic * newgfx = g_sdl_graphic_cache.getResizedGraphic(gfx, width, height);
    if (newgfx) {
        newgfx->blit(pimpl->screen, bl_x, (pimpl->screen->h - 1) - bl_y - (newgfx->getHeight() - 1));
    }
}

void Display::drawGraphicWithCC(int bl_x, int bl_y, const Graphic &gfx, int width, int height, const ColourChange &cc)
{
    const Graphic * newgfx = g_sdl_graphic_cache.getResizedGraphicWithCC(gfx, width, height, cc);
    if (newgfx) {
        newgfx->blit(pimpl->screen, bl_x, (pimpl->screen->h - 1) - bl_y - (newgfx->getHeight() - 1));
    }
}


//
// Pixel drawing
//

void Display::beginPixels()
{
    SDL_LockSurface(pimpl->screen);
}

void Display::drawPixel(int x, int y, int R, int G, int B)
{
    y = pimpl->screen->h - 1 - y;
    PlotPixel(pimpl->screen, x, y, R, G, B);
}

void Display::endPixels()
{
    SDL_UnlockSurface(pimpl->screen);
}
