/*
 * qconfig.cpp
 *
 * Copyright (c) Stephen Thompson, 2007.
 * Licensed for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "qconfig.hpp"

#include <cctype>
#include <istream>
#include <map>
using namespace std;

namespace {

	// This is where we store the varname/value pairs:
	map<string,int> intmap;
	map<string,float> floatmap;
	map<string,string> stringmap;

	// Eat up white space characters from stream, stop before reading
	// first non-WS character or at eof.
	void EatSpace(istream &str)
	{
		char c;
		while (str.get(c)) {
			if (c == '#') {
				// read comment
				while (str.get(c)) {
					if (c == '\n') break;
				}
			}
			if (!isspace(c)) {
				str.putback(c);
				break;
			}
		}
	}

	// Read a variable name string. Stop before reading first whitespace 
	// character or first equals sign.
	// Return true if nonempty varname read, else false.
	bool ReadVarName(istream &str, string &s)
	{
		char c;
		s.clear();
		while (str.get(c)) {
			if (isspace(c) || c=='=') {
				str.putback(c);
				break;
			}
			s += c;
		}
		return (!s.empty());
	}

	// Eat whitespace followed by a single equals sign.
	// Returns true if successful, or false if a non-ws char or EOF found
	// before seeing an equals sign.
	bool EatEquals(istream &str)
	{
		char c;
		while (str.get(c)) {
			if (c == '=') return true;
			if (c == '#') {
				// read comment
				while (str.get(c)) {
					if (c == '\n') break;
				}
			}
			if (!isspace(c)) {
				// error: other nonspace character
				str.putback(c);
				return false;
			}
		}
		// error: eof
		return false;
	}

	// Read a value string. Stop before reading first whitespace character.
	// If first non-whitespace character is a double quote, then treat it as a quoted string;
	// keep reading until the next double quote is found.
	// Returns false if unmatched " found, or if no value was read; else true.
    bool ReadValueString(istream &str, string &s)
	{
		char c;
		s.clear();

		str.get(c);
		if (c == '"') {
			// Quoted string: Keep reading until second " found, then stop.
			s += c;
			while (str.get(c)) {
				s += c;
				if (c=='"') return true;
			}
			return false;  // error: EOF met before second "
			
		} else {
			// Non quote found, so just keep going until first ws (or EOF)
			str.putback(c);
			while (str.get(c)) {
				if (isspace(c)) {
					str.putback(c);
					break;
				} else {
					s += c;
				}
			}
			return (!s.empty());
		}
	}


	void AddToTable(const string &varname, const string &value)
	{
		// Don't allow same variable to be added twice.
		if (intmap.find(varname) != intmap.end()
		|| floatmap.find(varname) != floatmap.end()
		|| stringmap.find(varname) != stringmap.end()) {
			throw QConfigLoadError(string("variable ") + varname + string(" defined twice"));
		}
		
		// Attempt to parse value
		// NB the checks for int and float could probably be improved :)
		// At the moment the parse always succeeds, possible adding a 0 value
		ASSERT(value.length() > 0);
		if (value[0]=='"') {
			ASSERT(value[value.length()-1] == '"');
			stringmap.insert(make_pair(varname, value.substr(1, value.length()-2)));
		} else if (value=="true") {
			intmap.insert(make_pair(varname, 1));
		} else if (value=="false") {
			intmap.insert(make_pair(varname, false));
		} else if (value.find('.') != string::npos
				   || value.find('e') != string::npos
				   || value.find('E') != string::npos) {
			// float
			floatmap.insert(make_pair(varname, atof(value.c_str())));
		} else {
			// int
			intmap.insert(make_pair(varname, atoi(value.c_str())));
		}
	}

}

//
// The load function
//

void QConfig::Load(istream &str)
{
	string s;
	string varname;
	string value;
	bool success;

	while (1) {
		EatSpace(str);
		success = ReadVarName(str, varname);
		if (!success) {
			// No more variables in the file!
			break;
		}

		// Try to read a value for this variable
		EatSpace(str);
		success = EatEquals(str);
		EatSpace(str);
		success = success && ReadValueString(str, value);
		if (success) {
			AddToTable(varname, value);
		} else {
			// Problem reading the value of this variable
			throw QConfigLoadError(string("error while reading variable \"") + varname + '"');
		}
	}
}


// 
// The get functions
//

bool QConfig::getBool(const string &name)
{
	// this is a bit of a cheat ...
	return bool(QConfig::getInt(name));
}

int QConfig::getInt(const string &name)
{
	map<string,int>::const_iterator i = intmap.find(name);
	if (i != intmap.end()) {
		return i->second;
	}

	map<string,float>::const_iterator f = floatmap.find(name);
	map<string,string>::const_iterator s = stringmap.find(name);
	if (f != floatmap.end() || s != stringmap.end()) {
		throw QConfigTypeMismatch(name);
	}

	throw QConfigNotFound(name);
}

float QConfig::getFloat(const string &name)
{
	map<string,float>::const_iterator f = floatmap.find(name);
	if (f != floatmap.end()) {
		return f->second;
	}

	map<string,int>::const_iterator i = intmap.find(name);
	if (i != intmap.end()) {
		return float(i->second);
	}

	map<string,string>::const_iterator s = stringmap.find(name);
	if (s != stringmap.end()) {
		throw QConfigTypeMismatch(name);
	}

	throw QConfigNotFound(name);
}

const string &QConfig::getString(const string &name)
{
	map<string,string>::const_iterator s = stringmap.find(name);
	if (s != stringmap.end()) {
		return s->second;
	}

	map<string,int>::const_iterator i = intmap.find(name);
	map<string,float>::const_iterator f = floatmap.find(name);
	if (i != intmap.end() || f != floatmap.end()) {
		throw QConfigTypeMismatch(name);
	}

	throw QConfigNotFound(name);
}


//
// Exceptions
//

const char * QConfigLoadError::what() const throw()
{
	return (string("Config File Error: ") + n).c_str();
}

const char * QConfigTypeMismatch::what() const throw()
{
	return (string("Config Variable \"") + n + string("\" has the wrong type")).c_str();
}

const char * QConfigNotFound::what() const throw()
{
	return (string("Config Variable \"") + n + string("\" not found!")).c_str();
}


